
/*
 * LIB/CONFIG.C	- Scan configuration file
 *
 * (c)Copyright 1998, Matthew Dillon, All Rights Reserved.  Refer to
 *    the COPYRIGHT file in the base directory of this distribution
 *    for specific rights granted.
 *
 */

#include "defs.h"

Prototype void LoadDiabloConfig(int ac, char **av);
Prototype void PrintVersion(void);

Prototype struct DiabloOpts DOpts;

int SetHashMethod(const char *opt, int *pmethod);
int SetCacheDirs(const char *opt, CacheDirType *cachedir);

struct DiabloOpts DOpts;

void
PrintVersion(void)
{
    printf("DIABLO Version %s - %s\n", VERS, SUBREV);
    exit(0);
}
void
initDOpts(void)
{
    /*
     * Set all the default options for the global config
     */
    memset(&DOpts, 0, sizeof(DOpts));
    DOpts.HashMethod = HASH_CRC;
    DOpts.CompatHashMethod = HASH_CRC;
    DOpts.HashSize = 16 * 1024 * 1024;
    DOpts.ReaderForks = 10;
    DOpts.ReaderThreads = 40;
    DOpts.ReaderFeedForks = 2;
    DOpts.ReaderDns = 5;
    DOpts.ReaderCacheMode = 1;
    DOpts.ReaderXOverMode = 1;
    DOpts.ReaderAutoAddToActive = 0;
    DOpts.FeederAutoAddToActive = 0;
    DOpts.ReaderDetailLog = 1;
    DOpts.RememberDays = 7;
    DOpts.HostCacheRebuildTime = 60 * 60;
    DOpts.DisplayAdminVersion = 1;
    DOpts.FeederRTStats = RTSTATS_NONE;
    DOpts.FeederXRefHost = NULL;
    DOpts.ReaderXRefHost = NULL;
    DOpts.ReaderXRefSlaveHost = NULL;
    DOpts.FeederPathHost = NULL;
    DOpts.FeederFilter = NULL;
    DOpts.ReaderPathHost = NULL;
    DOpts.NewsAdmin = NULL;
    DOpts.FeederHostName = NULL;
    DOpts.ReaderHostName = NULL;
    DOpts.NewsMaster = NULL;
    DOpts.ReaderCrashHandler = "none";
    DOpts.PathList = NULL;
    DOpts.FeederBindHost = NULL;
    DOpts.FeederPort = "nntp";
    DOpts.ReaderBindHost = NULL;
    DOpts.ReaderPort = "nntp";
    DOpts.ReaderCacheDirs.dt_dirlvl = 1;
    DOpts.ReaderCacheDirs.dt_dirinfo[0] = 512;
}

void
LoadDiabloConfig(int ac, char **av)
{
    FILE *fi;
    char buf[256];
    int versOk = 0;
    int exitMe = 0;
    int i;
    const char *diabloConfig;

    /*
     * Figure out path to diablo.config.  Check environment first, then
     * look for -C option to override it.  If neither is given, use the
     * default.
     */

    if ((diabloConfig = getenv("DIABLO_CONFIG_PATH")) == NULL) {
        diabloConfig = "%s/diablo.config";
    }

    for (i = 1; i < ac; ++i) {
	char *ptr = av[i];
	if (ptr[0] == '-' && ptr[1] == 'C') {
	    diabloConfig = ptr[2] ? ptr + 2 : av[++i];
	}
    }
    if (diabloConfig == NULL) {
	fprintf(stderr, "Expected argument to -C option\n");
	exit(1);
    }

    initDOpts();

    if ((fi = xfopen("r", diabloConfig, NewsHome)) != NULL) {
	while (fgets(buf, sizeof(buf), fi) != NULL) {
	    char *cmd;
	    char *opt = NULL;
	    int cmdErr = 1;

	    if (buf[0] == '\n' || buf[0] == '#')
		continue;
	    if ((cmd = strtok(buf, " \t\n")) != NULL &&
		(opt = strtok(NULL, " \t\n")) != NULL
	    ) {
		int optErr = 1;

		cmdErr = 0;
		if (strcasecmp(cmd, "version") == 0) {
		    optErr = 0;
		    if (strtod(opt, NULL) > strtod(VERS, NULL))
			versOk = 0;
		    else
			versOk = 1;
		} else if (versOk == 0) {
		    optErr = 0;
		} else if (strcasecmp(cmd, "expire") == 0) {
			/* Now unused */
		} else if (strcasecmp(cmd, "hash") == 0) {
		    if (opt) {
			char *p2;
			if ((p2 = strchr(opt, '/')) != NULL)
			    *p2++ = 0;
			optErr = SetHashMethod(opt, &DOpts.HashMethod);
			if (p2)
			    optErr = SetHashMethod(p2, &DOpts.CompatHashMethod);
		    }
		} else if (strcasecmp(cmd, "hsize") == 0) {
		    if (opt) {
			int n = strtol(opt, &opt, 0);

			switch(*opt) {
			case 'm':
			case 'M':
			    n *= 1024;
			    /* fall through */
			case 'k':
			case 'K':
			    n *= 1024;
			    optErr = 0;
			    break;
			default:
			    break;
			}
			if (n < 256 * 1024 || n > 128 * 1024 * 1024) {
			    optErr = 1;
			    fprintf(stderr, "Illegal Hash size: %d\n", n);
			    syslog(LOG_CRIT, "Illegal Hash size: %d", n);
			}
			if ((n ^ (n - 1)) != (n << 1) - 1) {
			    fprintf(stderr, "Hash size not a power of 2: %d\n", n);
			    syslog(LOG_CRIT, "Hash size not a power of 2: %d", n);
			    optErr = 1;
			}
			if (optErr == 0)
			    DOpts.HashSize = n;
		    }
		} else if (strcasecmp(cmd, "remember") == 0) {
		    if (opt) {
			DOpts.RememberDays = strtol(opt, NULL, 0);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "maxconnect") == 0) {
		    if (opt) {
			DOpts.MaxPerRemote = strtol(opt, NULL, 0);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "hostcacherebuildtime") == 0) {
		    if (opt) {
			DOpts.HostCacheRebuildTime = strtol(opt, NULL, 0);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "active") == 0) {
		    if (opt) {
			if (strcasecmp(opt, "on") == 0) {
			    optErr = 0;
			    DOpts.FeederActiveEnabled = 1;
			} else if (strcasecmp(opt, "off") == 0) {
			    optErr = 0;
			    DOpts.FeederActiveEnabled = 0;
			} else if (opt[0] == '/') {
			    optErr = 0;
			    DOpts.FeederActiveEnabled = 1;
			    ServerDActivePat = strdup(opt);
			}
		    }
		} else if (strcasecmp(cmd, "activedrop") == 0) {
		    if (opt) {
			if (strcasecmp(opt, "on") == 0) {
			    optErr = 0;
			    DOpts.FeederActiveDrop = 1;
			} else if (strcasecmp(opt, "off") == 0) {
			    optErr = 0;
			    DOpts.FeederActiveDrop = 0;
			}
		    }
		} else if (strcasecmp(cmd, "maxspool") == 0) {
		    /* No longer used */
		} else if (strcasecmp(cmd, "precommittime") == 0) {
		    if (opt) {
			SetPCommitExpire(atoi(opt), -1, -1);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "postcommittime") == 0) {
		    if (opt) {
			SetPCommitExpire(-1, atoi(opt), -1);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "hiscachesize") == 0) {
		    if (opt) {
			SetPCommitExpire(-1, -1, atoi(opt));
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "slave") == 0 ||
				strcasecmp(cmd, "feederxrefslave") == 0) {
		    if (opt) {
			if (strcasecmp(opt, "on") == 0) {
			    optErr = 0;
			    DOpts.FeederXRefSlave = 1;
			} else if (strcasecmp(opt, "off") == 0) {
			    optErr = 0;
			    DOpts.FeederXRefSlave = 0;
			}
		    }
		} else if (strcasecmp(cmd, "feederxrefsync") == 0) {
		    if (opt) {
			if (strcasecmp(opt, "on") == 0) {
			    optErr = 0;
			    DOpts.FeederXRefSync = 1;
			} else if (strcasecmp(opt, "off") == 0) {
			    optErr = 0;
			    DOpts.FeederXRefSync = 0;
			}
		    }
		} else if (strcasecmp(cmd, "displayadminversion") == 0) {
		    if (opt) {
			if (strcasecmp(opt, "on") == 0) {
			    optErr = 0;
			    DOpts.DisplayAdminVersion = 1;
			} else if (strcasecmp(opt, "off") == 0) {
			    optErr = 0;
			    DOpts.DisplayAdminVersion = 0;
			}
		    }
		} else if (strcasecmp(cmd, "autoaddtoactive") == 0) {
		    if (opt) {
			if (strcasecmp(opt, "on") == 0) {
			    optErr = 0;
			    DOpts.FeederAutoAddToActive = 1;
			    DOpts.ReaderAutoAddToActive = 1;
			} else if (strcasecmp(opt, "off") == 0) {
			    optErr = 0;
			    DOpts.FeederAutoAddToActive = 0;
			    DOpts.ReaderAutoAddToActive = 0;
			}
		    }
		} else if (strcasecmp(cmd, "feederautoaddtoactive") == 0) {
		    if (opt) {
			if (strcasecmp(opt, "on") == 0) {
			    optErr = 0;
			    DOpts.FeederAutoAddToActive = 1;
			} else if (strcasecmp(opt, "off") == 0) {
			    optErr = 0;
			    DOpts.FeederAutoAddToActive = 0;
			}
		    }
		} else if (strcasecmp(cmd, "readerautoaddtoactive") == 0) {
		    if (opt) {
			if (strcasecmp(opt, "on") == 0) {
			    optErr = 0;
			    DOpts.ReaderAutoAddToActive = 1;
			} else if (strcasecmp(opt, "off") == 0) {
			    optErr = 0;
			    DOpts.ReaderAutoAddToActive = 0;
			}
		    }
		} else if (strcasecmp(cmd, "readerdetaillog") == 0) {
		    if (opt) {
			if (strcasecmp(opt, "on") == 0) {
			    optErr = 0;
			    DOpts.ReaderDetailLog = 1;
			} else if (strcasecmp(opt, "off") == 0) {
			    optErr = 0;
			    DOpts.ReaderDetailLog = 0;
			}
		    }
		} else if (strcasecmp(cmd, "readerdns") == 0) {
		    if (opt) {
			DOpts.ReaderDns = strtol(opt, NULL, 0);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "readerforks") == 0) {
		    if (opt) {
			DOpts.ReaderForks = strtol(opt, NULL, 0);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "readerfeedforks") == 0) {
		    if (opt) {
			DOpts.ReaderFeedForks = strtol(opt, NULL, 0);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "readerthreads") == 0) {
		    if (opt) {
			DOpts.ReaderThreads = strtol(opt, NULL, 0);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "readercache") == 0) {
		    if (opt) {
			if (strcasecmp(opt, "on") == 0) {
			    optErr = 0;
			    DOpts.ReaderCacheMode = 1;
			} else if (strcasecmp(opt, "off") == 0) {
			    optErr = 0;
			    DOpts.ReaderCacheMode = 0;
			}
		    }
		} else if (strcasecmp(cmd, "readerxover") == 0) {
		    if (opt) {
			if (strcasecmp(opt, "on") == 0) {
			    DOpts.ReaderXOverMode = 1;
			    optErr = 0;
			} else if (strcasecmp(opt, "off") == 0) {
#if 0
			    DOpts.ReaderXOverMode = 0;	/* NOT YET SUPPORTED */
			    optErr = 0;
#endif
			} else if (strcasecmp(opt, "trackonly") == 0) {
			    DOpts.ReaderXOverMode = 2;
			    optErr = 0;
			}
		    }
		} else if (strncasecmp(cmd, "path_", 5) == 0) {
		    /*
		     * Paths (defaults already preset in globals)
		     */
		    const char **pptr = NULL;

		    if (strcasecmp(cmd + 5, "home") == 0)
			pptr = &NewsHome;
		    else if (strcasecmp(cmd + 5, "spool") == 0) 
			pptr = &SpoolHomePat;
		    else if (strcasecmp(cmd + 5, "dqueue") == 0) 
			pptr = &DQueueHomePat;
		    else if (strcasecmp(cmd + 5, "group") == 0) 
			pptr = &GroupHomePat;
		    else if (strcasecmp(cmd + 5, "cache") == 0) 
			pptr = &CacheHomePat;
		    else if (strcasecmp(cmd + 5, "log") == 0) 
			pptr = &LogHomePat;
		    else if (strcasecmp(cmd + 5, "feeds") == 0) 
			pptr = &FeedsHomePat;
		    else if (strcasecmp(cmd + 5, "lib") == 0) 
			pptr = &LibHomePat;
		    else if (strcasecmp(cmd + 5, "db") == 0) 
			pptr = &DbHomePat;
		    else if (strcasecmp(cmd + 5, "server_dactive") == 0) 
			pptr = &ServerDActivePat;
		    else if (strcasecmp(cmd + 5, "reader_dactive") == 0) 
			pptr = &ReaderDActivePat;
		    else if (strcasecmp(cmd + 5, "diablo_socket") == 0) 
			pptr = &DiabloSocketPat;
		    else if (strcasecmp(cmd + 5, "dreader_socket") == 0) 
			pptr = &DReaderSocketPat;
		    else if (strcasecmp(cmd + 5, "dexpire") == 0) 
			pptr = &DExpireCtlPat;
		    else if (strcasecmp(cmd + 5, "dcontrol") == 0) 
			pptr = &DControlCtlPat;
		    else if (strcasecmp(cmd + 5, "dhistory") == 0) 
			pptr = &DHistoryPat;
		    else if (strcasecmp(cmd + 5, "dumphist") == 0) 
			pptr = &DumpHistPat;
		    else if (strcasecmp(cmd + 5, "diablo_hosts") == 0) 
			pptr = &DiabloHostsPat;
		    else if (strcasecmp(cmd + 5, "dserver_hosts") == 0) 
			pptr = &DServerHostsPat;
		    else if (strcasecmp(cmd + 5, "moderators") == 0) 
			pptr = &ModeratorsPat;
		    else if (strcasecmp(cmd + 5, "spam_cache") == 0) 
			pptr = &SpamCachePat;
		    else if (strcasecmp(cmd + 5, "pcommit_cache") == 0) 
			pptr = &PCommitCachePat;
		    else if (strcasecmp(cmd + 5, "dexpover_list") == 0)
			pptr = &DExpireOverListPat;
		    else if (strcasecmp(cmd + 5, "dhosts_cache") == 0)
			pptr = &DHostsCachePat;
		    else if (strcasecmp(cmd + 5, "dnewsfeeds") == 0) 
			pptr = &DNewsfeedsPat;
		    else if (strcasecmp(cmd + 5, "dnntpspool") == 0) 
			pptr = &DNNTPSpoolCtlPat;
		    else if (strcasecmp(cmd + 5, "distrib_pats") == 0) 
			pptr = &DistribDotPatsPat;
		    else if (strcasecmp(cmd + 5, "distributions") == 0) 
			pptr = &DistributionsPat;
		    else if (strcasecmp(cmd + 5, "incominglog") == 0) 
			pptr = &IncomingLogPat;
		    else if (strcasecmp(cmd + 5, "drincominglog") == 0) 
			pptr = &DRIncomingLogPat;
		    else if (strcasecmp(cmd + 5, "feedstats") == 0) 
			pptr = &DFeedStatsPat;
		    else if (strcasecmp(cmd + 5, "pathlog") == 0) 
			pptr = &FPathLogPat;
		    else if (strcasecmp(cmd + 5, "shutdown_cleanup") == 0) 
			pptr = &ShutdownCleanup;
		    else
			cmdErr = 1;

		    /*
		     * XXX check for zero or one '%s' in opt.
		     */
		    if (cmdErr == 0 && opt != NULL) {
			*pptr = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "spooldirs") == 0) {
		    /* No longer used */
		} else if (strcasecmp(cmd, "port") == 0) {
		    if (opt) {
			DOpts.FeederPort = strdup(opt);
			DOpts.ReaderPort = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "feederbindport") == 0) {
		    if (opt) {
			DOpts.FeederPort = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "feederbindhost") == 0) {
		    if (opt) {
			DOpts.FeederBindHost = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "readerbindport") == 0) {
		    if (opt) {
			DOpts.ReaderPort = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "readerbindhost") == 0) {
		    if (opt) {
			DOpts.ReaderBindHost = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "feederrtstats") == 0) {
		    if (opt) {
			if (strcasecmp(opt, "none") == 0)
			    DOpts.FeederRTStats = RTSTATS_NONE;
			else if (strcasecmp(opt, "label") == 0)
			    DOpts.FeederRTStats = RTSTATS_LABEL;
			else if (strcasecmp(opt, "hostname") == 0)
			    DOpts.FeederRTStats = RTSTATS_HOST;
			else
			    printf("Unkown feederrealtimestats option: '%s'\n",
									opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "internalfilter") == 0) {
		    if (opt) {
			DOpts.SpamFilterOpt = strtol(opt, NULL, 0);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "rejectartswithnul") == 0) {
		    if (opt) {
			DOpts.RejectArtsWithNul = strtol(opt, NULL, 0);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "newsmaster") == 0) {
		    if (opt) {
			DOpts.NewsMaster = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "readercrash") == 0) {
		    if (opt) {
			DOpts.ReaderCrashHandler = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "xrefhost") == 0) {
		    if (opt) {
			DOpts.FeederXRefHost = strdup(opt);
			DOpts.ReaderXRefHost = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "feederxrefhost") == 0) {
		    if (opt) {
			DOpts.FeederXRefHost = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "readerxrefhost") == 0) {
		    if (opt) {
			DOpts.ReaderXRefHost = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "readerxrefslavehost") == 0) {
		    if (opt) {
			DOpts.ReaderXRefSlaveHost = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "feederfilter") == 0) {
		    if (opt) {
			DOpts.FeederFilter = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "readercachedirs") == 0) {
		    if (opt) {
			optErr = SetCacheDirs(opt, &DOpts.ReaderCacheDirs);
			if (optErr != 0)
			    syslog(LOG_CRIT, "Illegal reader cache dir format\n");
		    }
		} else if (strcasecmp(cmd, "readerpath") == 0) {
		    if (opt) {
			if (strcmp(opt, "0") == 0)
			    DOpts.ReaderPathHost = strdup("");
			else
			    DOpts.ReaderPathHost = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "path") == 0) {
		    if (opt) {
			PathListType *pl = NULL;

			if (strcmp(opt, "0") == 0) {
			    DOpts.ReaderPathHost = strdup("");
			    DOpts.FeederPathHost = strdup("");
			} else {
			    DOpts.ReaderPathHost = strdup(opt);
			    DOpts.FeederPathHost = strdup(opt);
			}
			pl = zalloc(&SysMemPool, sizeof(PathListType));
			pl->pathent = DOpts.FeederPathHost;
			pl->pathtype = 1;
			pl->next = DOpts.PathList;
			DOpts.PathList = pl;
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "feederpath") == 0) {
		    if (opt) {
			PathListType *pl = NULL;

			if (strcmp(opt, "0") == 0)
			    DOpts.FeederPathHost = strdup("");
			else
			    DOpts.FeederPathHost = strdup(opt);
			pl = zalloc(&SysMemPool, sizeof(PathListType));
			pl->pathent = DOpts.FeederPathHost;
			pl->pathtype = 1;
			pl->next = DOpts.PathList;
			DOpts.PathList = pl;
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "commonpath") == 0) {
		    if (opt) {
			PathListType *pl = NULL;

			pl = zalloc(&SysMemPool, sizeof(PathListType));
			pl->pathent = strdup(opt);
			pl->pathtype = 2;
			pl->next = DOpts.PathList;
			DOpts.PathList = pl;
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "newsadmin") == 0) {
		    if (opt) {
			DOpts.NewsAdmin = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "hostname") == 0) {
		    if (opt) {
			DOpts.FeederHostName = strdup(opt);
			DOpts.ReaderHostName = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "feederhostname") == 0) {
		    if (opt) {
			DOpts.FeederHostName = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "readerhostname") == 0) {
		    if (opt) {
			DOpts.ReaderHostName = strdup(opt);
			optErr = 0;
		    }
		} else if (strcasecmp(cmd, "newsmaster") == 0) {
		    if (opt) {
			DOpts.NewsMaster = strdup(opt);
			optErr = 0;
		    }
		} else {
		    cmdErr = 1;
		    optErr = 0;
		}
		if (optErr) {
		    exitMe = 1;
		    fprintf(stderr, "Unknown diablo.config command: %s %s\n",
						cmd, ((opt) ? opt : ""));
		    syslog(LOG_CRIT, "Unknown diablo.config command: %s %s",
						cmd, ((opt) ? opt : ""));
		}
	    }
	    if (cmdErr && cmd) {
		exitMe = 1;
		fprintf(stderr, "Unknown diablo.config command: %s %s\n",
						cmd, ((opt) ? opt : ""));
		syslog(LOG_CRIT, "Unknown diablo.config command: %s %s",
						cmd, ((opt) ? opt : ""));
	    }
	}
	fclose(fi);
    } else {
	fprintf(stderr, "Unable to open diablo.config using pattern %s\n",
							diabloConfig);
	syslog(LOG_CRIT, "Unable to open diablo.config using pattern %s",
							diabloConfig);
	exitMe = 1;
    }
    if (exitMe) {
	fprintf(stderr, "critical error parsing diablo.config, exiting.\n");
	exit(1);
    }
}

int
SetHashMethod(const char *opt, int *pmethod)
{
    int optErr = 0;

    if (strcasecmp(opt, "prime") == 0) {
	*pmethod = HASH_PRIME;
    } else if (strcasecmp(opt, "crc") == 0) {
	*pmethod = HASH_CRC;
    } else if (strcasecmp(opt, "ocrc") == 0) {
	*pmethod = HASH_OCRC;
    } else {
	optErr = 1;
    }
    return(optErr);
}

int
SetCacheDirs(const char *opt, CacheDirType *cachedir)
{
    int v;
    char *p;

    cachedir->dt_dirlvl = 0;
    v = atoi(opt);
    if (v < 1)
	return(1);
    cachedir->dt_dirinfo[cachedir->dt_dirlvl] = v;
    cachedir->dt_dirlvl++;
    while ((p = strchr(opt, '/')) != NULL) {
	p++;
	opt = p;
	v = atoi(p);
	if (v < 1)
	    return(1);
	cachedir->dt_dirinfo[cachedir->dt_dirlvl] = v;
	cachedir->dt_dirlvl++;
	if (cachedir->dt_dirlvl > 9)
	    break;
    }
    return(0);
}

